#version 330
#extension GL_ARB_separate_shader_objects : enable





layout(location = 0) smooth in vec2 _fin_fragTexCoord;

struct SFragmentInput
{
	vec2 fragTexCoord;
	bool IsFrontFace;
};

layout(location = 0) out vec4 _fout_outColor;

struct SFragmentOutput
{
	vec4 outColor;
};

 uniform sampler2D Texture;
 uniform sampler2D BlueNoise;

layout( std140) uniform UToneMappingInfo
{
	vec4 Common;
	vec4 ExpGammaSat;
	vec4 Vignetting;
	vec4 VignetteColor;
} ToneMappingInfo;
vec3 toSRGB(vec3 linearRGB)
{
	vec3 higher = vec3(1.055, 1.055, 1.055) * pow(linearRGB, vec3(1.0/2.4, 1.0/2.4, 1.0/2.4)) - vec3(0.055, 0.055, 0.055);
	vec3 lower = linearRGB * vec3(12.92, 12.92, 12.92);
	return mix(higher, lower, step(linearRGB, vec3(0.0031308, 0.0031308, 0.0031308)));
}

vec3 toLinear(vec3 sRGB)
{
	vec3 higher = pow((sRGB + vec3(0.055, 0.055, 0.055))/vec3(1.055, 1.055, 1.055), vec3(2.4, 2.4, 2.4));
	vec3 lower = sRGB/vec3(12.92, 12.92, 12.92);
	return mix(higher, lower, step(sRGB, vec3(0.04045, 0.04045, 0.04045)));
}


void FragmentMain( in SFragmentInput  fInput,  out SFragmentOutput  fOutput  )
{



	vec2 uv = fInput.fragTexCoord;


	fOutput.outColor =  texture(Texture, uv) ;
	{
		vec3 vignetteColor =  ToneMappingInfo.VignetteColor .rgb;
		float intensityColor =  ToneMappingInfo.Vignetting .x;
		float intensityDesaturation =  ToneMappingInfo.Vignetting .y;
		float roundness =  ToneMappingInfo.Vignetting .z;
		float smoothness =  ToneMappingInfo.Vignetting .w;
		float aspectRatio =  ToneMappingInfo.Common .z;

		float greyIntensity = 0.2126 * fOutput.outColor.r + 0.7152 * fOutput.outColor.g + 0.0722 * fOutput.outColor.b;

		vec2 uvFactor = normalize(vec2(1.f, mix(1.f, aspectRatio, roundness)));
		vec2 uvNormalized = 2.f * uvFactor * (uv - 0.5f);
		float alpha = length(uvNormalized);
		float cursor = min(pow(alpha, 1.f / (1.e-3f + smoothness)), 1.f);

		float weightColor = 1.f - intensityColor * cursor;
		float weightDesaturation = 1.f - intensityDesaturation * cursor;

		fOutput.outColor.rgb = weightColor * (weightDesaturation * fOutput.outColor.rgb + (1.f - weightDesaturation) * greyIntensity) +
								(1.f - weightColor) * vignetteColor;
	}



	{
		float exposure =  ToneMappingInfo.ExpGammaSat .x;
		vec3 gammaInv = 1. /  ToneMappingInfo.ExpGammaSat .yyy;
		float saturation =  ToneMappingInfo.ExpGammaSat .z;

		vec3 colorCorrected = fOutput.outColor.rgb;

		float greyIntensity = 0.2126 * colorCorrected.r + 0.7152 * colorCorrected.g + 0.0722 * colorCorrected.b;
		colorCorrected = max(colorCorrected * saturation + greyIntensity * (1.f - saturation), vec3(0, 0, 0));

		colorCorrected = 1.f - exp(- exposure * colorCorrected);

		float lumaGammaSpace = dot(pow(colorCorrected, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114));

		colorCorrected = pow(colorCorrected, gammaInv);

		fOutput.outColor.rgb = colorCorrected;
		fOutput.outColor.a = lumaGammaSpace;

		if ( ToneMappingInfo.ExpGammaSat .w != 0)
		{
			float screenWidth =  ToneMappingInfo.Common .x;
			float screenHeight =  ToneMappingInfo.Common .y;
			vec2 coords = uv * vec2(screenWidth,screenHeight);
			vec3 col = fOutput.outColor.rgb;
			if (gammaInv.x == 1.0f)
				col = toSRGB(col);
			vec3 colFloor = floor(col * 254.999999f)/255.0f;
			vec3 colCeil = colFloor + 1.0f/255.0f;
			vec3 noise =  texture(BlueNoise, coords / 32.0f) .rgb;
			col = mix(colFloor, colCeil, step(noise, (col-colFloor)/(colCeil-colFloor)));
			if (gammaInv.x == 1.0f)
				col = toLinear(col);
			fOutput.outColor.rgb = col;
		}
	}
}

void main()
{
	SFragmentInput fInput;
	SFragmentOutput fOutput;

	fInput.fragTexCoord = _fin_fragTexCoord;
	fInput.IsFrontFace = gl_FrontFacing;
	FragmentMain(fInput, fOutput);
	_fout_outColor = fOutput.outColor;
}
